REBOL []

install-HTTPd-extension [
	name: 'mod-fastcgi 
	
	order: [
		url-to-filename	first
		set-mime-type	normal
		access-check	normal
		make-response	normal
		logging			last
	]
		
	dot: #"."
	seq-id: 1
	register-list: none
	
	on-started: func [svc /local s ctx][
		register-list: extapp-register/with 'fastcgi [
			on-response: func [port data err /local job ctx][
				if err [log/warn err]
				job: find port/jobs port/id
				ctx: service/client
				service/set-peer job/2
				service/on-task-done data job/3
				remove/part job 3
				service/set-peer ctx
			]
			on-error: func [data][log/error mold data]
		][
			client: id: none
		]
		false
	]
		
	; --- Decline if not fastcgi script	
	declined?: func [req][not find register-list req/handler]
	
	url-to-filename: func [req /local d? cfg domain ext][
		d?: declined? req
		cfg: req/cfg
		if empty? trim req/in/target [		;-- trim should be done when target is parsed
			foreach file to-block any [select cfg 'default []][
				req/in/file: rejoin [cfg/root-dir req/in/path file]				
				if req/file-info: info? req/in/file [
					req/in/target: form file
					if ext: find/last req/in/target dot [
						req/in/ext: to word! ext
						req/handler: select service/handlers req/in/ext
					]
					if d?: declined? req [return false]
				]
			]
		]
		either d? [
			false
		][
			req/in/file: rejoin [cfg/root-dir req/in/path req/in/target]
			true
		]
	]
	
	set-mime-type: func [req][
		if declined? req [return none]		
		req/out/mime: 'text/html
		true
	]
	
	access-check: func [req /local info mdate][
		; --- This phase is redefined to avoid Last-Modified header generation
		; --- and 404 errors from mod-static (inappropriate here)
		
		if declined? req [return none]		
		true
	]
	
	make-response: func [req /local port jobs][	
		if declined? req [return none]
		
		set [port jobs] extapp-get-info req/handler		
		port/id: seq-id
		port/client: service/client		;-- TBD: remove and replace with fcgi-job parameters
		repend jobs [seq-id service/client req]
		seq-id: seq-id // 65000 + 1	;-- TBD: reuse id to be more server friendly ???

		insert-port port make fcgi-job-class [
			url: join req/in/path req/in/target
			;info: req/in/path
			path: form to-local-file get-modes req/in/file 'full-path
			script: req/in/url
			query: req/in/arg
			if req/auth/type [
				auth-type: form req/auth/type
				user: req/auth/user
			]
			headers: req/in/headers
			if content: req/in/content [
				cnt-length: select req/in/headers 'Content-Length
				cnt-type: select req/in/headers 'Content-Type
			]
			if req/in/method <> 'GET [method: form req/in/method]
		]
		true
	]
	
	logging: func [req][
		none
	]
	
	words: []
]
